/*$Id: measure_slewrate.cc,v 26.82 2008/06/01 14:20:25 al Exp $ -*- C++ -*-
 * Copyright (C) 2008 Albert Davis
 * Author: Albert Davis <aldavis@gnu.org>
 *
 * This file is part of "Gnucap", the Gnu Circuit Analysis Package
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *------------------------------------------------------------------
 */
#include "globals.h"
#include "s__.h"
#include "m_wave.h"
#include "u_function.h"
/*--------------------------------------------------------------------------*/
namespace {
/*--------------------------------------------------------------------------*/
class MEASURE : public FUNCTION {
public:
  std::string eval(CS& Cmd, CARD_LIST* Scope)const
  {
    std::string probe_name;
    PARAMETER<double> before(BIGBIG);
    PARAMETER<double> after(-BIGBIG);
    PARAMETER<double> from;
    PARAMETER<double> to;

    unsigned here = Cmd.cursor();
    Cmd >> probe_name;
    WAVE* w = SIM::find_wave(probe_name);

    if (!w) {
      Cmd.reset(here);
    }else{
    }

    here = Cmd.cursor();
    do {
      ONE_OF
	|| Get(Cmd, "probe", &probe_name)
	|| Get(Cmd, "before", &before)
	|| Get(Cmd, "after",  &after)
	|| Get(Cmd, "from|initial", &from)
	|| Get(Cmd, "to|final",     &to)
	;
    }while (Cmd.more() && !Cmd.stuck(&here));

    if (!Cmd.skip1b(')')) {
      Cmd.warn(bWARNING, "need )");
    }else{
    }

    if (!w) {
      w = SIM::find_wave(probe_name);
    }else{
    }
    
    if (w) {
      before.e_val(BIGBIG, Scope);
      after.e_val(-BIGBIG, Scope);
      from.e_val(0., Scope);
      to.e_val(0., Scope);

      enum STAT {WAITING, READY, IN_RANGE, DONE} stat = WAITING;
      double start_time = BIGBIG;
      double stop_time = BIGBIG;
      double start_val = BIGBIG;
      double stop_val  = BIGBIG;
      WAVE::const_iterator begin = lower_bound(w->begin(), w->end(), DPAIR(after, -BIGBIG));
      WAVE::const_iterator end   = upper_bound(w->begin(), w->end(), DPAIR(before, BIGBIG));
      WAVE::const_iterator lower = begin;
      for (WAVE::const_iterator i = begin; i < end && stat != DONE; ++i) {
	double val = i->second;
	switch (stat) {
	case WAITING:
	  if (in_order(val, double(from), double(to))) {
	    stat = READY;
	    start_time = i->first;
	    start_val = i->second;
	  }else{
	    // still WAITING
	  }
	  break;
	case READY:
	  if (in_order(val, double(from), double(to))) {
	    // still READY
	    start_time = i->first;
	    start_val = i->second;
	  }else if (in_order(double(from), val, double(to))) {
	    stat = IN_RANGE;
	    WAVE::const_iterator upper = i;
	    double position = (from - lower->second) / (upper->second - lower->second);
	    start_time = lower->first + position * (upper->first - lower->first);
	    start_val = from;
	  }else if (in_order(double(from), double(to), val)) {
	    // not accurate, was never IN_RANGE
	    stat = DONE;
	    double position = (from - lower->second) / (i->second - lower->second);
	    start_time = lower->first + position * (i->first - lower->first);
	    start_val = from;
	    position = (to - lower->second) / (i->second - lower->second);
	    stop_time = lower->first + position * (i->first - lower->first);
	    stop_val = to;
	  }else{untested();
	    // something is wrong
	  }
	  break;
	case IN_RANGE:
	  if (in_order(val, double(from), double(to))) {
	    // false start
	    stat = READY;
	    start_time = i->first;
	    start_val = i->second;
	  }else if (in_order(double(from), val, double(to))) {
	    // still IN_RANGE
	  }else if (in_order(double(from), double(to), val)) {
	    stat = DONE;
	    double position = (to - lower->second) / (i->second - lower->second);
	    stop_time = lower->first + position * (i->first - lower->first);
	    stop_val = to;
	  }else{untested();
	    // something is wrong
	  }
	  break;
	case DONE:
	  break;
	};
	lower = i;
      }
      if (stat == DONE) {
	assert(stop_time > start_time);
	//return to_string((stop_val-start_val)/(stop_time-start_time));
	return "{(" + to_string(stop_val) + "-" + to_string(start_val) + ")/(" 
	  + to_string(stop_time) + "-" + to_string(start_time) + ")}";
      }else{
	return to_string(BIGBIG);
      }
    }else{
      throw Exception_No_Match(probe_name);
    }
  }
} p3;
DISPATCHER<FUNCTION>::INSTALL d3(&function_dispatcher, "ddt|slewrate", &p3);
/*--------------------------------------------------------------------------*/
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
